# Natural Language Toolkit: Unit Tests
#
# Copyright (C) 2001-2026 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
# URL: <https://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
Unit tests for the NLTK modules.  These tests are intended to ensure
that source code changes don't accidentally introduce bugs.
For instructions, please see:

../../web/dev/local_testing.rst

https://github.com/nltk/nltk/blob/develop/web/dev/local_testing.rst


"""
